from bolib3 import np

"""
A trivial example with no constraints and no dependencies between the upper and lower-level. 
minimise    (x1 - 1)^2 + (x2 + 2)^2 + (x3 - 3)^2 + (x4 - 4)^2 
subject to  y = argmin((y1 - 1)^2 + (y2 + 2)^2 + (y3 - 3)^2 + (y4 - 4)^2 )
"""

# Properties
name: str = 'basic_unconstrained'
category: str = 'miscellaneous'
subcategory: str = ''
datasets: list = []

# Feasible point
x0 = [1.0,  -2.0,  3.0, -4.0]
y0 = [1.0,  -2.0,  3.0, -4.0]

# Methods
def F(x, y, data=None):
    """
    Upper-level objective function
    (quadratic)
    """
    return (x[0]-1)**2 + (x[1]+2)**2 + (x[2]-3)**2 + (x[3]+4)**2


def G(x, y, data=None):
    """
    Upper-level inequality constraints
    (none)
    """
    return np.empty(0)


def H(x, y, data=None):
    """
    Upper-level equality constraints
    (none)
    """
    return np.empty(0)


def f(x, y, data=None):
    """
    Lower-level objective function
    (quadratic)
    """
    return (y[0]-1)**2 + (y[1]+2)**2 + (y[2]-3)**2 + (y[3]+4)**2 + 5.0


def g(x, y, data=None):
    """
    Lower-level inequality constraints
    (none)
    """
    return np.empty(0)


def h(x, y, data=None):
    """
    Lower-level equality constraints
    (none)
    """
    return np.empty(0)


def read_data(filepath=''):
    """
    If the bilevel program is parameterized by data, this function should
     provide code to read data file and return an appropriate python structure.
    """
    pass


def dimensions(key='', data=None):
    """
    If the argument 'key' is not specified, then:
     - a dictionary mapping variable/function names (str) to the corresponding dimension (int) is returned.
    If the first argument 'key' is specified, then:
     - a single integer representing the dimension of the variable/function with the name {key} is returned.
    """
    n = {
        "x": 4,  # Upper-level variables
        "y": 4,  # Lower-level variables
        "F": 1,  # Upper-level objective functions
        "G": 0,  # Upper-level inequality constraints
        "H": 0,  # Upper-level equality constraints
        "f": 1,  # Lower-level objective functions
        "g": 0,  # Lower-level inequality constraints
        "h": 0,  # Lower-level equality constraints
    }
    if key in n:
        return n[key]
    return n
